#include <stdio.h>
#include <stdlib.h>

#define SIZE 100

struct { int A[SIZE], K, B[SIZE]; } Foo;
int *PA, *PB;
int offset;

void dump(int stride) {
  int i;
  printf("A[%03d ~ %03d]: ", SIZE-stride, SIZE-1);
  for (i=SIZE-stride; i<SIZE; i++)
    printf("%03d ", Foo.A[i]);

  printf("[K = %03d] ", Foo.K);

  printf("B[%03d ~ %03d]: ", 0, stride-1);
  for (i=0; i<stride; i++)
    printf("%03d ", Foo.B[i]);
  puts("\n");
}

void reset() {
  int i;
  for (i=0; i<SIZE; i++)
    Foo.A[i] = 0;
  Foo.K = 0;
  for (i=0; i<SIZE; i++)
    Foo.B[i] = i;
  PA = Foo.A+offset;
  PB = Foo.B+offset;
}

// Things that should *never* alias
int noAlias01 (int a) {
  int i;
  for (i=0; i<SIZE; i++)
    Foo.A[i] = Foo.B[i] + a;
  return Foo.A[a];
}


// Things that may alias (with vectors)
int mayAlias01 (int a) {
  int i;
  for (i=0; i<SIZE; i++)
    Foo.A[i] = Foo.B[SIZE-i-1] + a;
  return Foo.A[a];
}

int mayAlias02 (int a) {
  int i;
  for (i=0; i<SIZE; i++)
    Foo.A[SIZE-i-1] = Foo.B[i] + a;
  return Foo.A[a];
}

int mayAlias03 (int a) {
  int i;
  for (i=0; i<SIZE; i++)
    *(PA+i) = *(PB+SIZE-i-1) + a;
  return *(PA+a);
}


// Things that *must* alias
int mustAlias01 (int a) {
  int i;
  for (i=0; i<SIZE; i++)
    Foo.A[i+10] = Foo.B[SIZE-i-1] + a;
  return Foo.A[a];
}

int mustAlias02 (int a) {
  int i;
  for (i=0; i<SIZE; i++)
    Foo.A[i] = Foo.B[SIZE-i-10] + a;
  return Foo.A[a];
}

int mustAlias03 (int a) {
  int i;
  for (i=0; i<SIZE; i++)
    Foo.A[i+10] = Foo.B[SIZE-i-10] + a;
  return Foo.A[a];
}


int main(int argc, char** argv) {
  int K = 10;
  int ret = 0;
  if (argc > 1)
    K = atoi(argv[1]);
  if (argc > 2)
    offset = atoi(argv[2]);

  reset();
  ret += noAlias01(K);
  dump(10);

  reset();
  ret += mayAlias01(K);
  dump(10);
  reset();
  ret += mayAlias02(K);
  dump(10);
  reset();
  ret += mayAlias03(K);
  dump(10);

  reset();
  ret += mustAlias01(K);
  dump(10);
  reset();
  ret += mustAlias02(K);
  dump(10);
  reset();
  ret += mustAlias03(K);
  dump(10);

  return ret;
}
